{-----------------------------------------------------------------------------
 Unit Name: BackgroundThreads
 Project Name : Parametricas4
 Author:    Angel Montesinos Amilibia
 Dates: 2003-2009
 Purpose: Defines class TBackgroundThread that draws part of a bitmap with
    a color depending on the value of a function.

-----------------------------------------------------------------------------}
{

Parametricas4, workshop for the geometry of surfaces in R^4.


MIT License

Copyright (c) [2003 - 2018] [Angel Montesinos-Amilibia], email montesin@uv.es

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.


}


unit BackgroundThreads;

interface

uses
  Classes, Windows, Graphics, 
  VectorsAndUtil, FEvaluatorBytes_64_XE2, FunctionAndPartDerivs_64,
  FLinePalette, HigherDegreeLines;

type

  TColorArray = array[0..30000] of TColor;
  pTColorArray = ^TColorArray;

  TBackgroundThread = class(TThread)
  private
    xu, xv, xuu, xuv, xvv, xuuu, xuuv, xuvv, xvvv, t, Jt, vH, vB, vC: Vector4;
    cu, cv, qh, qv: Double;
    width, height: Integer;
    hh, hb, hc, bb, bc, cc, nh, jbc: Double;
    function AngleHEllipse(const p: VectorsAndUtil.Vector2) : Double;
    procedure ComputeBackgroundColor(fR2R1: MapR2R1);
    function ComputeHBCandProducts(p : VectorsAndUtil.Vector2): Boolean;
    function ComputeStrongPrincipalPolynomial(var p5: RealPolynomial; p: VectorsAndUtil.Vector2): Boolean;
    function DepthOfOrigin(const p: VectorsAndUtil.Vector2) : Double;
    function DiscriminantExtremalNormDevDirs(const p: VectorsAndUtil.Vector2): Double;
    function DiscriminantExtremalLatDevDirs(const p: VectorsAndUtil.Vector2): Double;
    function DiscriminantStrongPrincipalLines(const p: VectorsAndUtil.Vector2): Double;
    {function EllipseArea(const p: VectorsAndUtil.Vector2): Double;} {not used}
    function GaussCurvature(const p: VectorsAndUtil.Vector2): Double;
    function JBdotC(const p: VectorsAndUtil.Vector2) : Double;
    procedure PosFirstAndSecondDerivativesFR2R4(u, uve: Double;
             var y, yu, yv, yuu, yuv, yvv: Vector4);
    procedure PosFirstSecondThirdDerivativesFR2R4(u, v: Double);
  protected
    procedure Execute; override;
  public
    sds: FourEvaluators;  {functions that define the surface}
    spds: FourFPD; {for partial derivatives}
    bm: TBitmap;  {Tbitmap where the drawing is done}
    firstRow, lastRow, tag: Integer; {first and last of the rows of 'bm' that must be drawn}
    bckgPlt: Plt; {color palette that defines the colors to use}
    ff: MapR2R1; {the function that will be drawn in a color scale}
    constructor CreateBT(CreateSuspended: Boolean; cu0, cv0, qh0, qv0: Double;
         var bm0: TBitmap; {theFs: FourEvaluators; theFsPD: FourFPD;}
         firstRow0, lastRow0, tg: Integer);
  end;

var
  thr: array of TBackgroundThread; {will be so many threads as CPU in the machine}


implementation

uses
  SysUtils, Math,
  FParam4;

procedure Determinant(var valor : Double; mc : matrix44);
var
  i, j, k, imejor : Integer;
  m : matrix44;
  aVector : Vector4;
  aux : Double;
begin
  valor := 1;
  m := mc;
  for k := 1 to 2 do
    begin
      iMejor := 0;
      aux := 0;
      for i := k to 4 do
        if abs(m[i, k]) > aux then
          begin
            aux := abs(m[i, k]);
            iMejor := i;
          end;
      if imejor = 0 then
        begin
          valor := 0;
          Exit;
        end;
      if iMejor > k then
        begin
          aVector := m[iMejor];
          m[iMejor] := m[k];
          m[k] := aVector;
          valor := -valor * m[k, k];
        end
      else
        valor := valor * m[k, k];
      for j := k + 1 to 4 do
        m[k, j] := m[k, j] / m[k, k];
      for i := k + 1 to 4 do
        for j := k + 1 to 4 do
          m[i, j] := m[i, j] - m[k, j] * m[i, k];
    end;
  valor := valor * (m[3, 3] * m[4, 4] - m[3, 4] *
    m[4, 3]);
end;


function Invert(c: TColor): TColor;
{for passing colors directly to a Tbitmap via 'Scanline', it is necessary
 to change red by blue}
begin
  result:= (c and $000000FF) shl 16 + (c and $0000FF00) + (c and $00FF0000) shr 16;
end;


function TBackgroundThread.AngleHEllipse(const p: VectorsAndUtil.Vector2) : Double;
{Gives the absolute value of the sine of the angle between H and the elipse
 major axis}
var
  m: matrix44;
  aux, phi, si, co : Double;
  axis: Vector4;
  i: Integer;
begin
  Result := kTheError;
  if not ComputeHBCandProducts(p) then
    Exit;
  aux := hypot(bb - cc, 2 * bc);
  if (aux < 1.0E-8) then
    Result := kTheError
  else begin
    phi:= MyArctan2(2 * bc, bb - cc) / 2;
    sincos(phi,si,co);
    aux:= 0;
    for i:= 1 to 4 do begin
      axis[i]:= vB[i]*co + vC[i]*si;
      aux:= aux+ sqr(axis[i]);
    end;
    m[1]:= axis;
    m[2]:= t;
    m[3]:= Jt;
    m[4]:= vH;
    Determinant(result, m);
    aux:= sqrt(aux*hh);
    if aux > 1.0e-12 then
      result:= abs(result/aux)
    else
      result:= kTheError;
  end;
end;
  
procedure TBackgroundThread.ComputeBackgroundColor(fR2R1: MapR2R1);
{This is the routine that computes the function and draws the color. The
 palette has the color corresponding to zero value of the function at
 position 'kMedPltIndex'. Colors for positive values of the function go
 from 'kMedPltIndex' down to 1; colors for negative values go from
 from 'kMedPltIndex' up to 2*kMedPltIndex}
const
  cColorPitchAtZero = 2; {gives, at zero, the ratio of variation of the color
    number in the palette with respect to the variation of the function. The
    greater 'cColorPitchAtZero' the faster the color separates from the color
    assigned to the zero function value}
type
  TColorArray = array[0..10000] of TColor;
  pTColorArray = ^TColorArray;
var
  i, j : longint;
  p : VectorsAndUtil.Vector2;
  a, q: Double;
  indice: Integer;
  line: pTColorArray;
begin
  for j := firstRow to lastRow do begin
    if Terminated then
      Exit;
    p[2] := (qv - j) / cV; {pass from bitmap row to domain ordinate}
    line:= bm.Scanline[j];
    for i := 0 to width - 1 do begin
      p[1] := (i - qh) / cU; {pass from bitmap columnto domain abscissa}
      a:= fR2R1(p);  {value of the function}
      if a <> kTheError then begin
      {conversion of function value to color index in palette}
        a:= a*cColorPitchAtZero;

        {separation from the zero color in the palette. 'round(q)' ranges between
         0  for  'abs(a) = 0' and 'kMedPltIndex - 1' for 'abs(a) -> infinity'}
        q := (kMedPltIndex - 1) * abs(a) / (1 + abs(a));
        if a > 0 then
          indice := round(kMedPltIndex -1 - q)
        else
          indice := round(kMedPltIndex + q);
        line^[i] := Invert(bckgPlt[indice]); {draw the pixel}
      end
    end;
  end;
end;

function TBackgroundThread.ComputeHBCandProducts(p : VectorsAndUtil.Vector2): Boolean;
{Compute the basic quantities that define completely the curvature ellipse.
 See the memory "ComputationsOnSurfaces.pdf" in this same folder}
var
  x, G11, G12,  G21, G22: Vector4;
  E, F, G, aux, auxbis : Double;
  upuu, upuv, upvu, upvv, vpuu, vpuv, vpvu, vpvv : Double;
  i : Integer;
  m: matrix44;
begin
  Result := False;
  try
  PosFirstAndSecondDerivativesFR2R4(p[1], p[2], x, xu, xv, xuu, xuv, xvv);
  E := EuclInnerProd(xu, xu);
  F := EuclInnerProd(xu, xv);
  G := EuclInnerProd(xv, xv);
  auxbis := E * G - sqr(F);
  if auxbis > 1.0E-12 then
    begin
      aux := 1 / sqrt(E);
      auxbis := sqrt(auxbis);
      auxbis := aux / auxbis;
      for i := 1 to 4 do
        begin
          t[i] := xu[i] * aux;
          Jt[i] := auxbis * (E * xv[i] - F * xu[i]);
          G21[i] := E * xuv[i] - F * xuu[i];
        end;
    end
  else
    Exit;
  G11 := xuu;
  upuu := EuclInnerProd(xu, xuu);
  upuv := EuclInnerProd(xu, xuv);
  upvv := EuclInnerProd(xu, xvv);
  vpuu := EuclInnerProd(xv, xuu);
  vpuv := EuclInnerProd(xv, xuv);
  for i := 1 to 4 do
    begin
      G12[i] := E * xuv[i] - F * xuu[i] + 2 * upuu * xv[i] - (vpuu + upuv) *
        xu[i];
      G22[i] := E * (E * xvv[i] - F * xuv[i] + 2 * upuv * xv[i] - (vpuv + upvv)
        * xu[i]) - F * G12[i];
    end;
  upuu := upuu * aux;
  vpuu := EuclInnerProd(Jt, G11);
  upuv := EuclInnerProd(t, G12);
  vpuv := EuclInnerProd(Jt, G12);
  upvu := EuclInnerProd(t, G21);
  vpvu := EuclInnerProd(Jt, G21);
  upvv := EuclInnerProd(t, G22);
  vpvv := EuclInnerProd(Jt, G22);
  aux := aux * auxbis;
  auxbis := sqr(auxbis);
  for i := 1 to 4 do
    begin
      G11[i] := (G11[i] - upuu * t[i] - vpuu * Jt[i]) / E;
      G21[i] := (G21[i] - upvu * t[i] - vpvu * Jt[i]) * aux;
      G12[i] := (G12[i] - upuv * t[i] - vpuv * Jt[i]) * aux;
      G22[i] := (G22[i] - upvv * t[i] - vpvv * Jt[i]) * auxbis;
      vH[i] := (G11[i] + G22[i]) / 2;
      vB[i] := (G11[i] - G22[i]) / 2;
      vC[i] := (G12[i] + G21[i]) / 2;
    end;
  m[1]:= vC;
  m[2]:= t;
  m[3]:= Jt;
  m[4]:= vB;
  Determinant(jbc, m);

  hh := EuclInnerProd(vH, vH);
  nh:= sqrt(hh);
  bb := EuclInnerProd(vB, vB);
  cc := EuclInnerProd(vC, vC);
  bc := EuclInnerProd(vB, vC);
  hb := EuclInnerProd(vH, vB);
  hc := EuclInnerProd(vH, vC);
  result := True;
  except on exception do
    result:= False;
  end;
end;

function TBackgroundThread.ComputeStrongPrincipalPolynomial(var p5: RealPolynomial; p: VectorsAndUtil.Vector2): Boolean;
{Gives to the variable 'pol' the coefficients of the fifth-degree
 polynomial whose roots will give the strong principal directions at a
 point of the surface in R^4. See the memory
    "Asymp3orderInR5_StrongPpalDirsInR4.pdf"
 in this same folder. It it fails, results in False}
var
  raa, rab, rbb: Vector4;
  m: matrix44;
  E, F, G, d, uudu, uudv, uvdu, uvdv, vvdu, vvdv, duuuv, duuvv, duvvv, puu, quu,
    puv, quv, pvv, qvv: Double;
  i: Integer;
begin
    PosFirstSecondThirdDerivativesFR2R4(p[1], p[2]);
    E:= EuclEnergy(xu);
    F:= EuclInnerProd(xu,xv);
    G:= EuclEnergy(xv);
    d:= sqrt(abs(E*G-sqr(F)));
    result:= d > 0;
    if result then begin
      E:= E*d;
      F:= F*d;
      G:= G*d;
      for i:= 1 to 4 do begin
        raa[i]:= -F*xuu[i] + E*xuv[i];
        rab[i]:= -G*xuu[i]+ E*xvv[i];
        rbb[i]:= F*xvv[i] - G*xuv[i];
      end;
      m[1]:= xu;
      m[2]:= xv;
      m[3]:= xuu;
      m[4]:= xuv;
      duuuv:= determinant44(m);
      m[4]:= xvv;
      duuvv:= determinant44(m);
      m[3]:= xuv;
      duvvv:= determinant44(m);
      uudu:= EuclInnerProd(xuu,xu);
      uudv:= EuclInnerProd(xuu,xv);
      uvdu:= EuclInnerProd(xuv,xu);
      uvdv:= EuclInnerProd(xuv,xv);
      vvdu:= EuclInnerProd(xvv,xu);
      vvdv:= EuclInnerProd(xvv,xv);
      puu:= d*(G*uudu - F*uudv);
      quu:= d*(E*uudv - F*uudu);
      puv:= d*(G*uvdu - F*uvdv);
      quv:= d*(E*uvdv - F*uvdu);
      pvv:= d*(G*vvdu - F*vvdv);
      qvv:= d*(E*vvdv - F*vvdu);
      p5.degree:= 5;

      m[3]:= raa;
      m[4]:= xuuu;
      p5.coefs[5]:= determinant44(m) + 3*(F*quu +E*puu)*duuuv;

      m[4]:= xuuv;
      p5.coefs[4]:= 3*determinant44(m) + 3*(F*(2*quv+puu)+2*E*puv + G*quu)*duuuv +
             3*(E*puu+F*quu)*duuvv;
      m[3]:= rab;
      m[4]:= xuuu;
      p5.coefs[4]:= p5.coefs[4] + determinant44(m);

      m[4]:= xuuv;
      p5.coefs[3]:= 3*determinant44(m) + 3*(F*(2*puv+qvv)+2*G*quv + E*pvv)*duuuv +
             3*(2*(E*puv + F*quv)+F*puu + G*quu)*duuvv+
             3*(F*quu + E*puu)*duvvv;
      m[3]:= rbb;
      m[4]:= xuuu;
      p5.coefs[3]:= p5.coefs[3] + determinant44(m);
      m[3]:= raa;
      m[4]:= xuvv;
      p5.coefs[3]:= p5.coefs[3] + 3*determinant44(m);

      m[3]:= rbb;
      m[4]:= xuuv;
      p5.coefs[2]:= 3*determinant44(m) + 3*(F*(2*quv+puu)+2*E*puv + G*quu)*duvvv +
             3*(2*(G*quv + F*puv)+F*qvv + E*pvv)*duuvv+
             3*(F*pvv + G*qvv)*duuuv;
      m[3]:= rab;
      m[4]:= xuvv;
      p5.coefs[2]:= p5.coefs[2] + 3*determinant44(m);
      m[3]:= raa;
      m[4]:= xvvv;
      p5.coefs[2]:= p5.coefs[2] + determinant44(m);

      m[3]:= rab;
      p5.coefs[1]:= determinant44(m) + 3*(F*(2*puv+qvv)+2*G*quv + E*pvv)*duvvv +
             3*(G*qvv+F*pvv)*duuvv;
      m[3]:= rbb;
      m[4]:= xuvv;
      p5.coefs[1]:= p5.coefs[1] + 3*determinant44(m);

      m[3]:= rbb;
      m[4]:= xvvv;
      p5.coefs[0]:= determinant44(m) + 3*(F*pvv +G*qvv)*duvvv;
    end
end;

constructor TBackgroundThread.CreateBT(CreateSuspended: Boolean;
         cu0, cv0, qh0, qv0: Double;
         var bm0: TBitmap; {theFs: FourEvaluators;}
         firstRow0, lastRow0, tg: Integer);
var
  i: Integer;
begin
  inherited Create(CreateSuspended);
  bm:= bm0;
  sds:= theFsClones[tg];
  spds:= theFPDClones[tg];
  for i:= 0 to kLastPltColor do {create private clone of palette}
    bckgPlt[i]:= bckgPal[i];
  firstRow:= firstRow0;
  lastRow:= lastRow0;
  width:= bm0.Width;
  height:= bm0.Height;
  cu:= cu0;
  cv:= cv0;
  qh:= qh0;
  qv:= qv0;
  case param4.backgroundType of
    kColorBy0InEllipse:
      ff:=  DepthOfOrigin;
    kColorByBxC:
      ff:= JBdotC;
    kColorByAngleHE:
      ff:= AngleHEllipse;
    kColorByGK:
      ff:= GaussCurvature;
    kDiscrNormDev:
       ff:= DiscriminantExtremalNormDevDirs;
    kDiscrLatDev:
       ff:= DiscriminantExtremalLatDevDirs;
    kDiscrStrPpal:
       ff:= DiscriminantStrongPrincipalLines;
    else
      ff:= nil;
  end;
end;

function TBackgroundThread.DepthOfOrigin(const p: VectorsAndUtil.Vector2) : Double;
{Gives the "depth" of the origin of the normal sbspace in the ellipse}
begin
  Result := kTheError;
  if ComputeHBCandProducts(p) then begin
    result:= bb*cc-sqr(bc)- hypot(hb*cc-hc*bc, hc*bb-hb*bc);
  end;
end;


function TBackgroundThread.DiscriminantExtremalNormDevDirs(const p: VectorsAndUtil.Vector2) : Double;
{Gives the value of the discriminant of the polynomial equation whose roots give
 the extremal normal deviation directions}
var
  a0, a1, a2, a3, a4, a02, a12, a22, a32, a42: Double;
begin
  Result := kTheError;
  if ComputeHBCandProducts(p) then begin
    a0:= bc + hc;
    a1:= 2*(cc - bb - hb);
    a2:= -6*bc;
    a3:= 2*(bb - cc - hb);
    a4:= bc - hc;
    a02:= sqr(a0);
    a12:= sqr(a1);
    a22:= sqr(a2);
    a32:= sqr(a3);
    a42:= sqr(a4);

    result:= -27* a42*sqr(a12) - 4*a3*a1*a32*a12 + 18* a1*a2*a3*a4*a12+
             a22*a32*a12 + 144* a0*a2*a12*a42 - 4* a2*a22*a4*a12- 6*a0*a32*a4*a12
             + 18*a0*a1*a2*a3*a32 - 192*a02*a1*a3*a42 - 80*a0*a1*a22*a3*a4-
             27* a02*sqr(a32) + 256*a0*a02*a4*a42 - 4*a0*a2*a22*a32-
             128*a02*a22*a42 + 16*a0*sqr(a22)*a4 + 144*a02*a2*a32*a4;
  end;
end;

function TBackgroundThread.DiscriminantExtremalLatDevDirs(const p: VectorsAndUtil.Vector2) : Double;
{Gives the value of the discriminant of the polynomial equation whose roots give
 the extremal lateral deviation directions}
var
  a0, a1, a2, a3, a4, a02, a12, a22, a32, a42: Double;
begin
  Result := kTheError;
  if ComputeHBCandProducts(p) then begin
    a0:= cc-bb-hb;
    a1:= -2*(4*bc + hc);
    a2:= -6*(cc-bb);
    a3:= 2*(4*bc - hc);
    a4:= hb + cc - bb;
    a02:= sqr(a0);
    a12:= sqr(a1);
    a22:= sqr(a2);
    a32:= sqr(a3);
    a42:= sqr(a4);

    result:= -27* a42*sqr(a12) - 4*a3*a1*a32*a12 + 18* a1*a2*a3*a4*a12+
             a22*a32*a12 + 144* a0*a2*a12*a42 - 4* a2*a22*a4*a12- 6*a0*a32*a4*a12
             + 18*a0*a1*a2*a3*a32 - 192*a02*a1*a3*a42 - 80*a0*a1*a22*a3*a4-
             27* a02*sqr(a32) + 256*a0*a02*a4*a42 - 4*a0*a2*a22*a32-
             128*a02*a22*a42 + 16*a0*sqr(a22)*a4 + 144*a02*a2*a32*a4;
  end;
end;

function TBackgroundThread.DiscriminantStrongPrincipalLines(const p: VectorsAndUtil.Vector2):
  Double;
var
  a0, a1, a2, a3, a4, a5: Double;
  a: array[0..5,2..5] of Double;
  pl: RealPolynomial;
  i, j: Integer;
begin
  if ComputeStrongPrincipalPolynomial(pl, p) then begin
    a0:= pl.coefs[0]; a[0,2]:= sqr(a0);
    a1:= pl.coefs[1]; a[1,2]:= sqr(a1);
    a2:= pl.coefs[2]; a[2,2]:= sqr(a2);
    a3:= pl.coefs[3]; a[3,2]:= sqr(a3);
    a4:= pl.coefs[4]; a[4,2]:= sqr(a4);
    a5:= pl.coefs[5]; a[5,2]:= sqr(a5);
    for i:= 0 to 5 do begin
      for j:= 3 to 5 do
        a[i,j]:= a[i,j-1]*pl.coefs[i];
    end;
    result:=a[4,2]*(a[3,2]*(a[1,2]*a[2,2]-4*a[1,3]*a3
       +18*a0*a1*a2*a3 -a0*(4*a[2,3]+27*a0*a[3,2]))
       +2*(-2*a[2,3]*(a[1,2]-4*a0*a2)+a1*a2*(9*a[1,2]-40*a0*a2)*a3
       -3*a0*(a[1,2]-24*a0*a2)*a[3,2])*a4-(27*a[1,4]-144*a0*a[1,2]*a2
       +128*a[0,2]*a[2,2]+192*a[0,2]*a1*a3)*a[4,2]+256*a[0,3]*a[4,3])
       +2*(2*a[3,3]*(-a[1,2]*a[2,2]+4*a[1,3]*a3-18*a0*a1*a2*a3+a0*(4*a[2,3]+27*a0*a[3,2]))
       +a3*(9*a[2,3]*(a[1,2]-4*a0*a2)+2*a1*a2*(-20*a[1,2]+89*a0*a2)*a3
       +3*a0*(4*a[1,2]-105*a0*a2)*a[3,2])*a4
       +(-3*a[1,3]*a[2,2]+72*a[1,4]*a3-373*a0*a[1,2]*a2*a3+280*a[0,2]*a[2,2]*a3
       +6*a0*a1*(2*a[2,3]+85*a0*a[3,2]))*a[4,2]-2*a0*(9*a[1,3]-40*a0*a1*a2+400*a[0,2]*a3)*a[4,3])*a5
       +(16*a[1,3]*a3*(9*a[2,2]+10*a0*a4)-64*a[1,4]*(2*a[3,2]+3*a2*a4)
       -10*a0*a1*a3*(63*a[2,3]+90*a0*a[3,2]+205*a0*a2*a4)
       +a[1,2]*(-27*a[2,4]+560*a0*a2*a[3,2]+1020*a0*a[2,2]*a4-50*a[0,2]*a[4,2])
       +a0*(108*a[2,5]+825*a0*a[2,2]*a[3,2]-900*a0*a[2,3]*a4
       +2250*a[0,2]*a[3,2]*a4+2000*a[0,2]*a2*a[4,2]))*a[5,2]
       +2*(128*a[1,5]-800*a0*a[1,3]*a2+1000*a[0,2]*a[1,2]*a3
       -1875*a[0,3]*a2*a3+125*a[0,2]*a1*(9*a[2,2]-10*a0*a4))*a[5,3]
       +3125*a[0,4]*a[5,4];
  end
  else
    result:= kTheError;
end;

{function TBackgroundThread.EllipseArea(const p: VectorsAndUtil.Vector2): Double;
begin
  Result := kTheError;
  if not ComputeHBCandProducts(p) then
    Exit;
  Result := sqrt(abs(bb * cc - sqr(bc)));
end;}

procedure TBackgroundThread.Execute;
begin
  if Assigned(ff) then
    ComputeBackgroundColor(ff);
end;

function TBackgroundThread.GaussCurvature(const p: VectorsAndUtil.Vector2) : Double;
begin
  Result := kTheError;
  if not ComputeHBCandProducts(p) then
    Exit;
  result:= hh - cc - bb;
end;

function TBackgroundThread.JBdotC(const p: VectorsAndUtil.Vector2): Double;
{Gives the inner product of 'JB' with 'C'. It vanishes when 'B' and 'C'
 are linearly dependent, therefore when the ellipse degenerates to a
 segment}
begin
  Result := kTheError;
  if not ComputeHBCandProducts(p) then
    Exit;
  result:= jbc;
end;

procedure TBackgroundThread.PosFirstAndSecondDerivativesFR2R4(u, uve: Double;
    var y, yu, yv, yuu, yuv, yvv: Vector4);
  var
    fd: SecondDeriv;
    i: Integer;
begin
  for i:= 1 to 4 do begin
    fd:= sds[i].SecondD([u,uve,1,0,0,0]);
    y[i]:= fd.fa0;
    yu[i]:= fd.dfa0;
    yuu[i]:= fd.ddfa0;
    fd:= sds[i].SecondD([u,uve,0,1,0,0]);
    yv[i]:= fd.dfa0;
    yvv[i]:= fd.ddfa0;
    fd:= sds[i].SecondD([u,uve,1,1,0,0]);
    yuv[i]:= 0.5*(fd.ddfa0 - yuu[i]- yvv[i]);
  end;
end;

procedure TBackgroundThread.PosFirstSecondThirdDerivativesFR2R4(u, v:
  Double);
{gives the value of the map at (u,v) and its first, second and third derivatives}
var
  x: Vector4;
  i: Integer;
begin
  PosFirstAndSecondDerivativesFR2R4(u, v, x, xu, xv, xuu, xuv, xvv);
  for i := 1 to 4 do begin
     xuuu[i]:= spds[i].ff[1,1,1]([u,v]);
     xuuv[i]:= spds[i].ff[1,1,2]([u,v]);
     xuvv[i]:= spds[i].ff[1,2,2]([u,v]);
     xvvv[i]:= spds[i].ff[2,2,2]([u,v]);
  end;
 end;
end.
