{-----------------------------------------------------------------------------
 Unit Name: FParametrizationR2R4
 Project Name : Parametricas4
 Author:    Angel Montesinos Amilibia
 Dates: 2003-2009
 Purpose: Implements class 'TParametrizationR2R4' that collects several computations
      related to surfaces in R^4.
-----------------------------------------------------------------------------}
{
Parametricas4, workshop for the geometry of surfaces in R^4.


MIT License

Copyright (c) [2003 - 2018] [Angel Montesinos-Amilibia], email montesin@uv.es

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.


}


{See the paper "ComputationsOnSurfaces.pdf" in this same folder}

unit FParametrizationR2R4;
interface
uses
  Windows, Graphics,
  FEvaluatorBytes_64_XE2, VectorsAndUtil;

type
  TParametrizationR2R4 = class
    t, Jt, x, xu, xv, xuv, xuu, xvv, xuuu, xvvv, xuuv, xuvv, vH, vB, vC:
      Vector4;
    G11, G12, G21, G22: Vector4;
    guu, guv, gvv, nb, nc, nh, bb, cc, bc, hb, hc, hh, jhb, jhc, jbc,
      smallValue: Double;
    function AngleHEllipse(const p: Vector2): Double;
    function ComputeCross(const par: vector2; const cClass: Integer;
      var halfMajorAxis, halfMinorAxis, longH, depth, alpha, KGAuss,
      discrNormDev, discrLatDev, discrStrPpal: Double): Boolean;
    {function ComputeCuspFunct(const p: Vector2): Double;}
    {function ComputeCuspFromDerivatives: Double;}
    procedure ComputeFff(const p: vector2);
    function ComputeHBC(p: vector2): Boolean;
    //function ComputeHBCandThirdDerivatives(p: vector2): Boolean;
    function ComputeHBCandProducts(p: vector2): Boolean;
    function ComputeInvariants(const par: vector2; var halfMajorAxis,
      halfMinorAxis, longH, depth, alpha, KGauss, discrNormDev, discrLatDev,
      discrStrPpal: Double): Boolean;
    {function ComputeInvariantsAndCusp(const par: vector2; var halfMajorAxis,
      halfMinorAxis, longH, depth, alpha, KGAuss, discrNormDev, discrLatDev: Double): Boolean;}
    {function ComputePQR(x, y: Double; var p, q, r, a, b, c: Double):
      Boolean;}
    function DepthOfOrigin(const p: Vector2): Double;
    function Directions(const par: vector2; const curveClass: Integer;
      var theTangent: vector2; var phi: Double): Byte;
    function DiscriminantExtremalNormDevDirs(const p: Vector2): Double;
    function DiscriminantExtremalLatDevDirs(const p: Vector2): Double;
    function DiscriminantStrongPrincipalLines(const p: Vector2): Double;
    procedure DrawOrthogonalVector(can: TCanvas; param: vector2; firstPoint,
      lastPoint: TPoint);
    function EllipseArea(const p: Vector2): Double;
    function GaussCurvature(const p: Vector2): Double;
    function JBdotC(const p: Vector2): Double;
    function JHdotB(const p: Vector2): Double;
    procedure PosFirstDerivativesFR2R4(u, v: Double; var y, yu, yv:
      Vector4);
    procedure PosFirstAndSecondDerivativesFR2R4(u, v: Double; var
      y, yu, yv, yuu, yuv, yvv: Vector4);
    procedure PosFirstSecondThirdDerivativesFR2R4(u, v: Double);
  end;



var
  MapR2R4: TParametrizationR2R4;

implementation
uses
  Classes, SysUtils, Math, Forms,
  FParam4, HigherDegreeLines, FunctionAndPartDerivs_64;

const
  eMinScale = 1.0E-12;
  eMinRootsSep = 1.0E-5;
  eSmallShift = 1.0E-8;

procedure Determinant(var valor: Double; mc: matrix44);
{determinant of the matrix 'mc'}
var
  i, j, k, imejor: Integer;
  m: matrix44;
  aVector: Vector4;
  aux: Double;
begin
  valor:= 1;
  m:= mc;
  for k:= 1 to 2 do begin
    iMejor:= 0;
    aux:= 0;
    for i:= k to 4 do
      if abs(m[i, k]) > aux then begin
        aux:= abs(m[i, k]);
        iMejor:= i;
      end;
    if imejor = 0 then begin
      valor:= 0;
      Exit;
    end;
    if iMejor > k then begin
      aVector:= m[iMejor];
      m[iMejor]:= m[k];
      m[k]:= aVector;
      valor:= -valor * m[k, k];
    end else
      valor:= valor * m[k, k];
    for j:= k + 1 to 4 do
      m[k, j]:= m[k, j] / m[k, k];
    for i:= k + 1 to 4 do
      for j:= k + 1 to 4 do
        m[i, j]:= m[i, j] - m[k, j] * m[i, k];
  end;
  valor:= valor * (m[3, 3] * m[4, 4] - m[3, 4] *
    m[4, 3]);
end;

procedure Finalize;
begin
  FreeAndNil(MapR2R4);
end;

function fmod(x, y: Double): Double;
var
  a: Double;
begin
  if abs(y) < 1.0E-30 then
    Result:= y
  else begin
    a:= trunc(x / y);
    result:= x - a * y;
  end;
end;

procedure Initialize;
begin
  MapR2R4:= TParametrizationR2R4.Create;
end;

function TParametrizationR2R4.AngleHEllipse(const p: Vector2): Double;
{See comments to 'ComputeInvariants'}
var
  m: matrix44;
  aux, phi, si, co: Double;
  axis: Vector4;
  i: Integer;
begin
  Result:= kTheError;
  if not ComputeHBCandProducts(p) then
    Exit;
  aux:= hypot(bb - cc, 2 * bc);
  if (aux < 1.0E-8) then
    Result:= kTheError
  else begin
    phi:= MyArctan2(2 * bc, bb - cc) / 2;
    sincos(phi, si, co);
    aux:= 0;
    for i:= 1 to 4 do begin
      axis[i]:= vB[i] * co + vC[i] * si;
      aux:= aux + sqr(axis[i]);
    end;
    m[1]:= axis;
    m[2]:= t;
    m[3]:= Jt;
    m[4]:= vH;
    Determinant(result, m);
    aux:= sqrt(aux) * nh;
    if aux > 1.0e-12 then
      result:= result / aux
    else
      result:= kTheError;
  end;
end;

function TParametrizationR2R4.ComputeCross(const par: vector2;
  const cClass: Integer;
  var halfMajorAxis, halfMinorAxis, longH, depth, alpha, KGauss,
  discrNormDev, discrLatDev, discrStrPpal: Double):
  Boolean;
{Computes the directions of the foliations of class 'cClass' at the
 domain point 'par'. The directions are returned in 'crosses'. As a byproduct,
 gives also the values of some invariants at 'par'. See doc of ComputeInvariants
 below. For the meaning of these directions, see the doc to the procedure
 'Directions' below and "ComputationsOnSurfaces.pdf" in this same folder}
var
  angles: Roots; {Here angles.values[i,1] denotes the angle of the i-th root,
  except when the computation of the roots are made in the
  unit ExtremalDevLines.pas; then it denotes the sine of that angle
  and angles.values[i,2] the cosine}
  dir1, dir2, p1, p2, p3, p4: vector2;
  axis: Vector4;
  m: Matrix44;
  a, b, c, g, p, q, aux, phi_M, si, co, side, gg: Double;
  i, j: Integer;
begin
  Result:= False;
  if not ComputeHBCandProducts(par) then
    Exit;
  try
    discrNormDev:= DiscriminantExtremalNormDevDirs(par);
    discrLatDev:= DiscriminantExtremalLatDevDirs(par);
    discrStrPpal:= DiscriminantStrongPrincipalLines(par);
    a:= 1 / sqrt(guu);
    aux:= a / sqrt(abs(guu * gvv - sqr(guv)));
    b:= -guv * aux;
    c:= guu * aux;

    KGauss:= hh - bb - cc;
    aux:= bb * cc - sqr(bc);
    if abs(aux) < 1.0E-12 then
      depth:= -1
    else
      depth:= 1 - hypot(hb * cc - hc * bc, hc * bb - hb * bc) / aux;
    longH:= sqrt(abs(hh));
    g:= sqrt(sqr(bb - cc) + sqr(2 * bc));
    gg:= bb + cc;
    halfMajorAxis:= sqrt(abs((gg + g) / 2));
    halfMinorAxis:= sqrt(abs((gg - g) / 2));
    if (g <= 1.0E-8) then
      alpha:= kTheError
    else begin
      phi_M:= MyArctan2(2 * bc, bb - cc) / 2;
      sincos(phi_M, si, co);
      aux:= 0;
      for i:= 1 to 4 do begin
        axis[i]:= vB[i] * co + vC[i] * si;
        aux:= aux + sqr(axis[i]);
      end;
      m[1]:= axis;
      m[2]:= t;
      m[3]:= Jt;
      m[4]:= vH;
      Determinant(alpha, m);
      aux:= sqrt(aux) * nh;
      if aux > 1.0e-12 then
        alpha:= abs(alpha / aux)
      else
        alpha:= kTheError;
    end;
    case cClass of
      kMeanCurvLines: begin
          if hypot(jhb, jhc) < 1.0e-8 then
            Exit;
          gg:= fmod(MyArctan2(-jhb, jhc) + 2 * pi, pi) / 2;
          co:= cos(gg);
          side:= sin(gg);
          dir1[1]:= a * co + b * side;
          dir1[2]:= c * side;
          p:= co;
          co:= -side;
          side:= p;
          dir2[1]:= a * co + b * side;
          dir2[2]:= c * side;
          aux:= hypot(dir1[1], dir1[2]);
          if aux < 1.0e-12 then
            Exit;
          p:= 0.025 * Param4.span2D / aux;
          aux:= hypot(dir2[1], dir2[2]);
          if aux < 1.0e-12 then
            Exit;
          q:= 0.025 * Param4.span2D / aux;
          for i:= 1 to 2 do begin
            dir1[i]:= p * dir1[i];
            dir2[i]:= q * dir2[i];
            p1[i]:= par[i] - dir1[i];
            p2[i]:= par[i] + dir1[i];
            p3[i]:= par[i] - dir2[i];
            p4[i]:= par[i] + dir2[i];
          end;
        end;
      kAsymptoticLines: begin
          aux:= hypot(jhb, jhc);
          if aux < 1.0e-12 then
            Exit;
          gg:= jbc / aux;
          if abs(gg) > 1 then
            Exit;
          gg:= arcsin(gg);
          phi_M:= MyArcTan2(jhc, jhb);
          co:= cos((phi_M + gg) / 2);
          side:= sin((phi_M + gg) / 2);
          dir1[1]:= a * co + b * side;
          dir1[2]:= c * side;
          co:= cos((phi_M - gg + pi) / 2);
          side:= sin((phi_M - gg + pi) / 2);
          dir2[1]:= a * co + b * side;
          dir2[2]:= c * side;
          aux:= hypot(dir1[1], dir1[2]);
          if aux < 1.0e-12 then
            Exit;
          p:= 0.025 * Param4.span2D / aux;
          aux:= hypot(dir2[1], dir2[2]);
          if aux < 1.0e-12 then
            Exit;
          q:= 0.025 * Param4.span2D / aux;
          for i:= 1 to 2 do begin
            dir1[i]:= p * dir1[i];
            dir2[i]:= q * dir2[i];
            p1[i]:= par[i] - dir1[i];
            p2[i]:= par[i] + dir1[i];
            p3[i]:= par[i] - dir2[i];
            p4[i]:= par[i] + dir2[i];
          end;
        end;
      kCurvatureLines: if (g > 1.0e-8) then begin
          co:= cos(phi_M / 2);
          side:= sin(phi_M / 2);
          dir1[1]:= a * co + b * side;
          dir1[2]:= c * side;
          p:= co;
          co:= -side;
          side:= p;
          dir2[1]:= a * co + b * side;
          dir2[2]:= c * side;
          aux:= hypot(dir1[1], dir1[2]);
          if aux < 1.0e-12 then
            Exit;
          p:= 0.025 * Param4.span2D / aux;
          aux:= hypot(dir2[1], dir2[2]);
          if aux < 1.0e-12 then
            Exit;
          q:= 0.025 * Param4.span2D / aux;
          for i:= 1 to 2 do begin
            dir1[i]:= p * dir1[i];
            dir2[i]:= q * dir2[i];
            p1[i]:= par[i] - dir1[i];
            p2[i]:= par[i] + dir1[i];
            p3[i]:= par[i] - dir2[i];
            p4[i]:= par[i] + dir2[i];
          end;
        end;
      kPrincipalLines: if (hypot(hc, hb) > 0) then begin
          phi_M:= fmod(2 * pi + MyArcTan2(hc, hb) / 2, pi / 2);
          co:= cos(phi_M);
          side:= sin(phi_M);
          dir1[1]:= a * co + b * side;
          dir1[2]:= c * side;
          p:= co;
          co:= -side;
          side:= p;
          dir2[1]:= a * co + b * side;
          dir2[2]:= c * side;
          aux:= hypot(dir1[1], dir1[2]);
          if aux < 1.0e-12 then
            Exit;
          p:= 0.025 * Param4.span2D / aux;
          aux:= hypot(dir2[1], dir2[2]);
          if aux < 1.0e-12 then
            Exit;
          q:= 0.025 * Param4.span2D / aux;
          for i:= 1 to 2 do begin
            dir1[i]:= p * dir1[i];
            dir2[i]:= q * dir2[i];
            p1[i]:= par[i] - dir1[i];
            p2[i]:= par[i] + dir1[i];
            p3[i]:= par[i] - dir2[i];
            p4[i]:= par[i] + dir2[i];
          end;
        end;
      kExtrNormDevLines, kExtrLatDevLines, kStrongPrincipalLines: begin
          GetHigherDegreeDirections(cClass, angles, par);
          Param4.curvesParts:= 2*angles.number;
          if angles.number > 0 then begin
            for i:= 1 to angles.number do begin
              if cClass = kStrongPrincipalLines then begin
                dir1[1]:= angles.values[i,1];
                dir1[2]:= angles.values[i,2];
              end
              else begin
                dir1[1]:= a*angles.values[i,1] + b*angles.values[i,2];
                dir1[2]:= c*angles.values[i,2];
              end;
              for j:= 1 to 2 do begin
                p1[j]:= par[j] - dir1[j];
                p2[j]:= par[j] + dir1[j];
              end;
              Param4.Cross[i, 1]:= Param4.FromR2toWindow(p1);
              Param4.Cross[i, 2]:= Param4.FromR2toWindow(p2);
            end;
          end;
        end;
    else
      ;
    end;
    if not (cClass in [kExtrNormDevLines, kExtrLatDevLines, kStrongPrincipalLines]) then  begin
      Param4.Cross[1, 1]:= Param4.FromR2toWindow(p1);
      Param4.Cross[1, 2]:= Param4.FromR2toWindow(p2);
      Param4.Cross[2, 1]:= Param4.FromR2toWindow(p3);
      Param4.Cross[2, 2]:= Param4.FromR2toWindow(p4);
    end;
    if (cClass = kCurvatureLines) and (g > 1.0e-8) then begin
      co:= cos(phi_M / 2 + pi / 4);
      side:= sin(phi_M / 2 + pi / 4);
      dir1[1]:= a * co + b * side;
      dir1[2]:= c * side;
      p:= co;
      co:= -side;
      side:= p;
      dir2[1]:= a * co + b * side;
      dir2[2]:= c * side;
      aux:= hypot(dir1[1], dir1[2]);
      if aux < 1.0e-12 then
        Exit;
      p:= 0.025 * Param4.span2D / aux;
      aux:= hypot(dir2[1], dir2[2]);
      if aux < 1.0e-12 then
        Exit;
      q:= 0.025 * Param4.span2D / aux;
      for i:= 1 to 2 do begin
        dir1[i]:= p * dir1[i];
        dir2[i]:= q * dir2[i];
        p1[i]:= par[i] - dir1[i];
        p2[i]:= par[i] + dir1[i];
        p3[i]:= par[i] - dir2[i];
        p4[i]:= par[i] + dir2[i];
      end;
      Param4.Cross[3, 1]:= Param4.FromR2toWindow(p1);
      Param4.Cross[3, 2]:= Param4.FromR2toWindow(p2);
      Param4.Cross[4, 1]:= Param4.FromR2toWindow(p3);
      Param4.Cross[4, 2]:= Param4.FromR2toWindow(p4);
    end;
    Result:= True;
  except on E: Exception do
      result:= False;
  end;
end;

procedure TParametrizationR2R4.ComputeFff(const p: vector2);
{computes and saves the coefficients of the first fundamental form }
var
  y, yu, yv: Vector4;
begin
  PosFirstDerivativesFR2R4(p[1], p[2], y, yu, yv);
  guu:= EuclEnergy(yu);
  guv:= EuclInnerProd(yu, yv);
  gvv:= EuclEnergy(yv);
end;

function TParametrizationR2R4.ComputeHBC(p: vector2): Boolean;
var
  x: Vector4;
  aux, auxbis: Double;
  upuu, upuv, upvu, upvv, vpuu, vpuv, vpvu, vpvv: Double;
  i: Integer;
begin
  Result:= True;
  try
    PosFirstAndSecondDerivativesFR2R4(p[1], p[2], x, xu, xv, xuu, xuv, xvv);
    guu:= EuclInnerProd(xu, xu);
    guv:= EuclInnerProd(xu, xv);
    gvv:= EuclInnerProd(xv, xv);
    auxbis:= guu * gvv - sqr(guv);
    if auxbis > 1.0E-12 then begin
      aux:= 1 / sqrt(guu);
      auxbis:= sqrt(auxbis);
      auxbis:= aux / auxbis;
      for i:= 1 to 4 do begin
        t[i]:= xu[i] * aux;
        Jt[i]:= auxbis * (guu * xv[i] - guv * xu[i]);
        G21[i]:= guu * xuv[i] - guv * xuu[i];
      end;
    end else
      Exit;
    G11:= xuu;
    upuu:= EuclInnerProd(xu, xuu);
    upuv:= EuclInnerProd(xu, xuv);
    upvv:= EuclInnerProd(xu, xvv);
    vpuu:= EuclInnerProd(xv, xuu);
    vpuv:= EuclInnerProd(xv, xuv);
    for i:= 1 to 4 do begin
      G12[i]:= guu * xuv[i] - guv * xuu[i] + 2 * upuu * xv[i] - (vpuu + upuv) *
        xu[i];
      G22[i]:= guu * (guu * xvv[i] - guv * xuv[i] + 2 * upuv * xv[i] - (vpuv +
        upvv)
        * xu[i]) - guv * G12[i];
    end;
    upuu:= upuu * aux;
    vpuu:= EuclInnerProd(Jt, G11);
    upuv:= EuclInnerProd(t, G12);
    vpuv:= EuclInnerProd(Jt, G12);
    upvu:= EuclInnerProd(t, G21);
    vpvu:= EuclInnerProd(Jt, G21);
    upvv:= EuclInnerProd(t, G22);
    vpvv:= EuclInnerProd(Jt, G22);
    aux:= aux * auxbis;
    auxbis:= sqr(auxbis);
    for i:= 1 to 4 do begin
      G11[i]:= (G11[i] - upuu * t[i] - vpuu * Jt[i]) / guu;
      G21[i]:= (G21[i] - upvu * t[i] - vpvu * Jt[i]) * aux;
      G12[i]:= (G12[i] - upuv * t[i] - vpuv * Jt[i]) * aux;
      G22[i]:= (G22[i] - upvv * t[i] - vpvv * Jt[i]) * auxbis;
      vH[i]:= (G11[i] + G22[i]) / 2;
      vB[i]:= (G11[i] - G22[i]) / 2;
      vC[i]:= (G12[i] + G21[i]) / 2;
    end;
  except on exception do
    result:= False;
  end;
end;

function TParametrizationR2R4.ComputeHBCandProducts(p: vector2): Boolean;
{Stores all basis data of the first and second fundamental forms}
var
  x: Vector4;
  aux, auxbis: Double;
  upuu, upuv, upvu, upvv, vpuu, vpuv, vpvu, vpvv: Double;
  i: Integer;
  m: matrix44;
begin
  Result:= False;
  try
    PosFirstAndSecondDerivativesFR2R4(p[1], p[2], x, xu, xv, xuu, xuv, xvv);
    guu:= EuclInnerProd(xu, xu);
    guv:= EuclInnerProd(xu, xv);
    gvv:= EuclInnerProd(xv, xv);
    auxbis:= guu * gvv - sqr(guv);
    if auxbis > 1.0E-12 then begin
      aux:= 1 / sqrt(guu);
      auxbis:= sqrt(auxbis);
      auxbis:= aux / auxbis;
      for i:= 1 to 4 do begin
        t[i]:= xu[i] * aux;
        Jt[i]:= auxbis * (guu * xv[i] - guv * xu[i]);
        G21[i]:= guu * xuv[i] - guv * xuu[i];
      end;
    end else
      Exit;
    G11:= xuu;
    upuu:= EuclInnerProd(xu, xuu);
    upuv:= EuclInnerProd(xu, xuv);
    upvv:= EuclInnerProd(xu, xvv);
    vpuu:= EuclInnerProd(xv, xuu);
    vpuv:= EuclInnerProd(xv, xuv);
    for i:= 1 to 4 do begin
      G12[i]:= guu * xuv[i] - guv * xuu[i] + 2 * upuu * xv[i] - (vpuu + upuv) *
        xu[i];
      G22[i]:= guu * (guu * xvv[i] - guv * xuv[i] + 2 * upuv * xv[i] - (vpuv +
        upvv)
        * xu[i]) - guv * G12[i];
    end;
    upuu:= upuu * aux;
    vpuu:= EuclInnerProd(Jt, G11);
    upuv:= EuclInnerProd(t, G12);
    vpuv:= EuclInnerProd(Jt, G12);
    upvu:= EuclInnerProd(t, G21);
    vpvu:= EuclInnerProd(Jt, G21);
    upvv:= EuclInnerProd(t, G22);
    vpvv:= EuclInnerProd(Jt, G22);
    aux:= aux * auxbis;
    auxbis:= sqr(auxbis);
    for i:= 1 to 4 do begin
      G11[i]:= (G11[i] - upuu * t[i] - vpuu * Jt[i]) / guu;
      G21[i]:= (G21[i] - upvu * t[i] - vpvu * Jt[i]) * aux;
      G12[i]:= (G12[i] - upuv * t[i] - vpuv * Jt[i]) * aux;
      G22[i]:= (G22[i] - upvv * t[i] - vpvv * Jt[i]) * auxbis;
      vH[i]:= (G11[i] + G22[i]) / 2;
      vB[i]:= (G11[i] - G22[i]) / 2;
      vC[i]:= (G12[i] + G21[i]) / 2;
    end;
    m[1]:= vB;
    m[2]:= t;
    m[3]:= Jt;
    m[4]:= vH;
    Determinant(jhb, m);

    m[1]:= vC;
    m[2]:= t;
    m[3]:= Jt;
    m[4]:= vH;
    Determinant(jhc, m);

    m[1]:= vC;
    m[2]:= t;
    m[3]:= Jt;
    m[4]:= vB;
    Determinant(jbc, m);

    hh:= EuclInnerProd(vH, vH);
    nh:= sqrt(hh);
    bb:= EuclInnerProd(vB, vB);
    nb:= sqrt(bb);
    cc:= EuclInnerProd(vC, vC);
    nc:= sqrt(cc);
    bc:= EuclInnerProd(vB, vC);
    hb:= EuclInnerProd(vH, vB);
    hc:= EuclInnerProd(vH, vC);
    result:= True;
  except on exception do
      result:= False;
  end;
end;

function TParametrizationR2R4.ComputeInvariants(const par: vector2; var
  halfMajorAxis, halfMinorAxis,
  longH, depth, alpha, KGauss, discrNormDev, discrLatDev, discrStrPpal: Double): Boolean;
{Gives at x(p) the values of the curvature ellipse half-axes, the length of the
 mean curvature vector H, the depth of the origin in the ellipse (negative values if
 it is out of te ellipse, at 'alpha' the sine of the angle that H forms with
 the major axis of the ellipse, the Gauss curvature, and the values of the
 discriminants of the curves of extremal normal or lateral deviation.}
var
  axis: Vector4;
  phi_M, a, g, aux, si, co: Double;
  m: matrix44;
  i: Integer;
begin
  Result:= False;
  if not ComputeHBCandProducts(par) then
    Exit;
  try
    discrStrPpal:= DiscriminantStrongPrincipalLines(par);
    discrNormDev:= DiscriminantExtremalNormDevDirs(par);
    discrLatDev:= DiscriminantExtremalLatDevDirs(par);
    KGauss:= hh - cc - bb;
    longH:= nh;
    aux:= hypot(bb - cc, 2 * bc);
    g:= bb + cc;
    halfMajorAxis:= sqrt(abs(g + aux) / 2);
    halfMinorAxis:= sqrt(abs(g - aux) / 2);
    if (bb + cc) < 1.0e-5 * hh then
      depth:= -nh
    else begin
      g:= 4 * (aux * (sqr(hb) + sqr(hc)) + (sqr(hb) - sqr(hc)) * (bb - cc) + 4 *
        hb * hc * bc) / (aux + bb + cc);
      a:= hh + aux;
      depth:= halfMajorAxis - sqrt(0.5 * (a + sqrt(abs(sqr(a) - g))));
    end;
    if (aux < 1.0E-8) then
      alpha:= kTheError
    else begin
      phi_M:= MyArctan2(2 * bc, bb - cc) / 2;
      sincos(phi_M, si, co);
      aux:= 0;
      for i:= 1 to 4 do begin {compute the major axis and its squared norm}
        axis[i]:= vB[i] * co + vC[i] * si;
        aux:= aux + sqr(axis[i]);
      end;
      m[1]:= axis;
      m[2]:= t;
      m[3]:= Jt;
      m[4]:= vH;
      Determinant(alpha, m);
      aux:= sqrt(aux) * nh;

      {dividing 'alpha' by the norms of 'vH' and  'axis' we get the sine of the angle
       between them}
      if aux > 1.0e-12 then
        alpha:= abs(alpha / aux)
      else
        alpha:= kTheError;
    end;
    Result:= True;
  except on E: Exception do
      result:= False;
  end;
end;

function TParametrizationR2R4.DepthOfOrigin(const p: Vector2): Double;
var
  aux, a, b: Double;
begin
  Result:= kTheError;
  if not ComputeHBCandProducts(p) then
    Exit;
  aux:= bb * cc - sqr(bc);
  a:= (hb * cc - hc * bc);
  b:= (hc * bb - hb * bc);
  Result:= aux - hypot(a, b);
end;

function TParametrizationR2R4.Directions(const par: vector2; const curveClass:
  Integer;
  var theTangent: vector2; var phi: Double): Byte;
{similar to ComputeCross}
const
  cShift = 0.001;
var
  a, b, c, aux, si, co, side, p, r: Double;
  ang: array[0..5] of Double;
  i, i0: Integer;
begin
  Result:= 0;
  if not ComputeHBCandProducts(par) then
    Exit;
  guu:= EuclInnerProd(xu, xu);
  guv:= EuclInnerProd(xu, xv);
  gvv:= EuclInnerProd(xv, xv);
  a:= 1 / sqrt(guu);
  aux:= a / sqrt(guu * gvv - sqr(guv));
  b:= -guv * aux;
  c:= guu * aux;
  case curveClass of
    kMeanCurvLines: begin
        if hypot(jhb, jhc) < 1.0e-8 then
          Exit;
        ang[2]:= fmod(MyArctan2(-jhb, jhc) + 2 * pi, pi) / 2;
        if phi = kTheError then
          ang[1]:= ang[2]
        else begin
          i:= system.trunc(2 * (phi - ang[2]) / pi);
          ang[1]:= ang[2] + i * pi / 2;
          ang[3]:= phi - ang[1];
          if ang[3] > pi / 4 then begin
            if (i < 3) then
              ang[1]:= ang[1] + pi / 2
            else
              ang[1]:= ang[2];
          end else if ang[3] < -pi / 4 then
            ang[1]:= ang[2] + 3 * pi / 2;
        end;
        phi:= ang[1];
        sincos(2 * phi + cshift, si, co);
        if jhb * co + jhc * si > 0 then
          result:= 1
        else
          result:= 2;
      end;
    kAsymptoticLines: begin
        aux:= hypot(jhb, jhc);
        if aux < 1.0e-7 then
          Exit;
        ang[3]:= jbc / aux;
        if abs(ang[3]) > 1 then
          Exit;
        ang[3]:= arcsin(ang[3]);
        ang[2]:= MyArcTan2(jhc, jhb);
        ang[1]:= (ang[2] + ang[3]) / 2 + pi;
        if ang[1] >= pi then
          ang[1]:= ang[1] - pi;
        ang[2]:= (ang[2] - ang[3] + pi) / 2;
        if ang[2] >= 2 * pi then
          ang[2]:= ang[2] - 2 * pi
        else if ang[2] >= pi then
          ang[2]:= ang[2] - pi
        else if ang[2] < 0 then
          ang[2]:= ang[2] + pi;
        if phi = kTheError then
          phi:= ang[1]
        else if phi = -kTheError then
          phi:= ang[2]
        else begin
          if ang[1] > ang[2] then begin
            ang[3]:= ang[1];
            ang[1]:= ang[2];
            ang[2]:= ang[3];
          end;
          ang[3]:= ang[1] + pi;
          ang[4]:= ang[2] + pi;
          ang[5]:= ang[1] + 2 * pi;
          ang[0]:= ang[4] - 2 * pi;
          r:= 100;
          i0:= -1;
          for i:= 0 to 5 do begin
            p:= abs(phi - ang[i]);
            if p < r then begin
              i0:= i;
              r:= p;
            end
          end;
          if i0 > -1 then begin
            if i0 = 5 then
              phi:= ang[1]
            else if i0 = 0 then
              phi:= ang[4]
            else
              phi:= ang[i0];
          end else
            Exit;
        end;
        sincos(2 * phi + cShift, si, co);
        if (jhb * si - jhc * co - jbc) > 0 then
          result:= 1
        else
          result:= 2;
      end;
    kCurvatureLines: begin
        aux:= sqrt(sqr(bb - cc) + sqr(2 * bc));
        if aux < 1.0E-12 then
          Exit;
        ang[1]:= (2 * pi + MyArctan2(2 * bc, bb - cc)) / 4;
        if phi = kTheError then
          ang[3]:= ang[1]
        else begin
          ang[1]:= fmod(phi, pi / 4);
          i:= system.trunc(4 * (phi - ang[1]) / pi);
          ang[3]:= ang[1] + i * pi / 4;
          ang[2]:= phi - ang[3];
          if ang[2] > pi / 8 then begin
            if (i < 7) then
              ang[3]:= ang[3] + pi / 4
            else
              ang[3]:= ang[1];
          end else if ang[2] < -pi / 8 then
            ang[3]:= ang[1] + 7 * pi / 4;
        end;
        phi:= ang[3];
        sincos(4 * phi + cShift, si, co);
        if 2 * bc * co - (bb - cc) * si > 0 then
          result:= 1
        else
          result:= 2;
      end;
    kPrincipalLines: begin
        ang[1]:= fmod(2 * pi + MyArcTan2(hc, hb) / 2, pi / 2);
        if phi = kTheError then
          ang[3]:= ang[1]
        else begin
          i:= system.trunc(2 * (phi - ang[1]) / pi);
          ang[3]:= ang[1] + i * pi / 2;
          ang[2]:= phi - ang[3];
          if ang[2] > pi / 4 then begin
            if (i < 3) then
              ang[3]:= ang[3] + pi / 2
            else
              ang[3]:= ang[1];
          end else if ang[2] < -pi / 4 then
            ang[3]:= ang[1] + 3 * pi / 2;
        end;
        phi:= ang[3];
        sincos(2 * phi + cShift, si, co);
        if hc * co - hb * si > 0 then
          result:= 1
        else
          result:= 2;
      end;
    kExtrNormDevLines:
        ;
    kExtrLatDevLines:
        ;
    kStrongPrincipalLines:
        ;
  else
    Exit;
  end;
  if phi > 2 * pi then
    beep;
  co:= cos(phi);
  side:= sin(phi);
  theTangent[1]:= a * co + b * side;
  theTangent[2]:= c * side;
end;

function TParametrizationR2R4.DiscriminantExtremalLatDevDirs(const p: Vector2):
  Double;
{Gives the value of the discriminant of the polynomial equation whose roots give
 the extremal lateral deviation directions}
var
  a0, a1, a2, a3, a4, a02, a12, a22, a32, a42: Double;
begin
  Result:= kTheError;
  if ComputeHBCandProducts(p) then begin
    a0:= cc - bb - hb;
    a1:= -2 * (4 * bc + hc);
    a2:= -6 * (cc - bb);
    a3:= 2 * (4 * bc - hc);
    a4:= hb + cc - bb;
    a02:= sqr(a0);
    a12:= sqr(a1);
    a22:= sqr(a2);
    a32:= sqr(a3);
    a42:= sqr(a4);

    result:= -27 * a42 * sqr(a12) - 4 * a3 * a1 * a32 * a12 + 18 * a1 * a2 * a3
      * a4 * a12 +
      a22 * a32 * a12 + 144 * a0 * a2 * a12 * a42 - 4 * a2 * a22 * a4 * a12 - 6
      * a0 * a32 * a4 * a12
      + 18 * a0 * a1 * a2 * a3 * a32 - 192 * a02 * a1 * a3 * a42 - 80 * a0 * a1
      * a22 * a3 * a4 -
      27 * a02 * sqr(a32) + 256 * a0 * a02 * a4 * a42 - 4 * a0 * a2 * a22 * a32
      -
      128 * a02 * a22 * a42 + 16 * a0 * sqr(a22) * a4 + 144 * a02 * a2 * a32 *
      a4;
  end;
end;


function TParametrizationR2R4.DiscriminantExtremalNormDevDirs(const p: Vector2):
  Double;
{Gives the value of the discriminant of the polynomial equation whose roots give
 the extremal normal deviation directions}
var
  a0, a1, a2, a3, a4, a02, a12, a22, a32, a42: Double;
begin
  Result:= kTheError;
  if ComputeHBCandProducts(p) then begin
    a0:= bc + hc;
    a1:= 2 * (cc - bb - hb);
    a2:= -6 * bc;
    a3:= 2 * (bb - cc - hb);
    a4:= bc - hc;
    a02:= sqr(a0);
    a12:= sqr(a1);
    a22:= sqr(a2);
    a32:= sqr(a3);
    a42:= sqr(a4);

    result:= -27 * a42 * sqr(a12) - 4 * a3 * a1 * a32 * a12 + 18 * a1 * a2 * a3
      * a4 * a12 +
      a22 * a32 * a12 + 144 * a0 * a2 * a12 * a42 - 4 * a2 * a22 * a4 * a12 - 6
      * a0 * a32 * a4 * a12
      + 18 * a0 * a1 * a2 * a3 * a32 - 192 * a02 * a1 * a3 * a42 - 80 * a0 * a1
      * a22 * a3 * a4 -
      27 * a02 * sqr(a32) + 256 * a0 * a02 * a4 * a42 - 4 * a0 * a2 * a22 * a32
      -
      128 * a02 * a22 * a42 + 16 * a0 * sqr(a22) * a4 + 144 * a02 * a2 * a32 *
      a4;
  end;
end;

function TParametrizationR2R4.DiscriminantStrongPrincipalLines(const p: Vector2):
  Double;
var
  a0, a1, a2, a3, a4, a5: Double;
  a: array[0..5,2..5] of Double;
  pl: RealPolynomial;
  i, j: Integer;
begin
  if ComputeStrongPrincipalPolynomial(pl, p) then begin
    a0:= pl.coefs[0]; a[0,2]:= sqr(a0);
    a1:= pl.coefs[1]; a[1,2]:= sqr(a1);
    a2:= pl.coefs[2]; a[2,2]:= sqr(a2);
    a3:= pl.coefs[3]; a[3,2]:= sqr(a3);
    a4:= pl.coefs[4]; a[4,2]:= sqr(a4);
    a5:= pl.coefs[5]; a[5,2]:= sqr(a5);
    for i:= 0 to 5 do begin
      for j:= 3 to 5 do
        a[i,j]:= a[i,j-1]*pl.coefs[i];
    end;
    result:=a[4,2]*(a[3,2]*(a[1,2]*a[2,2]-4*a[1,3]*a3
       +18*a0*a1*a2*a3 -a0*(4*a[2,3]+27*a0*a[3,2]))
       +2*(-2*a[2,3]*(a[1,2]-4*a0*a2)+a1*a2*(9*a[1,2]-40*a0*a2)*a3
       -3*a0*(a[1,2]-24*a0*a2)*a[3,2])*a4-(27*a[1,4]-144*a0*a[1,2]*a2
       +128*a[0,2]*a[2,2]+192*a[0,2]*a1*a3)*a[4,2]+256*a[0,3]*a[4,3])
       +2*(2*a[3,3]*(-a[1,2]*a[2,2]+4*a[1,3]*a3-18*a0*a1*a2*a3+a0*(4*a[2,3]+27*a0*a[3,2]))
       +a3*(9*a[2,3]*(a[1,2]-4*a0*a2)+2*a1*a2*(-20*a[1,2]+89*a0*a2)*a3
       +3*a0*(4*a[1,2]-105*a0*a2)*a[3,2])*a4
       +(-3*a[1,3]*a[2,2]+72*a[1,4]*a3-373*a0*a[1,2]*a2*a3+280*a[0,2]*a[2,2]*a3
       +6*a0*a1*(2*a[2,3]+85*a0*a[3,2]))*a[4,2]-2*a0*(9*a[1,3]-40*a0*a1*a2+400*a[0,2]*a3)*a[4,3])*a5
       +(16*a[1,3]*a3*(9*a[2,2]+10*a0*a4)-64*a[1,4]*(2*a[3,2]+3*a2*a4)
       -10*a0*a1*a3*(63*a[2,3]+90*a0*a[3,2]+205*a0*a2*a4)
       +a[1,2]*(-27*a[2,4]+560*a0*a2*a[3,2]+1020*a0*a[2,2]*a4-50*a[0,2]*a[4,2])
       +a0*(108*a[2,5]+825*a0*a[2,2]*a[3,2]-900*a0*a[2,3]*a4
       +2250*a[0,2]*a[3,2]*a4+2000*a[0,2]*a2*a[4,2]))*a[5,2]
       +2*(128*a[1,5]-800*a0*a[1,3]*a2+1000*a[0,2]*a[1,2]*a3
       -1875*a[0,3]*a2*a3+125*a[0,2]*a1*(9*a[2,2]-10*a0*a4))*a[5,3]
       +3125*a[0,4]*a[5,4];
  end
  else
    result:= kTheError;
end;

procedure TParametrizationR2R4.DrawOrthogonalVector(can: TCanvas; param:
  vector2; firstPoint,
  lastPoint: TPoint);
{compute and draw a vector orthogonal and of same norm as the vector
 defined by two positions of the mouse, 'firstPoint,lastPoint'}
var
  det: Double;
  sh, orth: vector2;
  raya: TPoint;
begin
  ComputeFff(param);
  sh[1]:= (lastPoint.x - firstPoint.x) * Param4.edges2D[1];
  sh[2]:= (firstPoint.y - lastPoint.y) * Param4.edges2D[2];
  det:= sqrt(abs(guu * gvv - sqr(guv)));
  if det < 1.0E-10 then Exit;
  orth[1]:= -(guv * sh[1] + gvv * sh[2]) / det;
  orth[2]:= (guu * sh[1] + guv * sh[2]) / det;
  raya.x:= firstPoint.x + round(orth[1] / Param4.edges2D[1]);
  raya.y:= firstPoint.y - round(orth[2] / Param4.edges2D[2]);
  can.moveto(firstPoint.x, firstPoint.y);
  can.lineto(raya.x, raya.y);
end;


function TParametrizationR2R4.EllipseArea(const p: Vector2): Double;
begin
  Result:= kTheError;
  if not ComputeHBCandProducts(p) then
    Exit;
  Result:= sqrt(abs(bb * cc - sqr(bc)));
end;

function TParametrizationR2R4.GaussCurvature(const p: Vector2): Double;
begin
  Result:= kTheError;
  if not ComputeHBCandProducts(p) then
    Exit;
  result:= hh - cc - bb;
end;

function TParametrizationR2R4.JBdotC(const p: Vector2): Double;
begin
  Result:= kTheError;
  if not ComputeHBCandProducts(p) then
    Exit;
  result:= jbc;
end;

function TParametrizationR2R4.JHdotB(const p: Vector2): Double;
begin
  Result:= kTheError;
  if not ComputeHBCandProducts(p) then
    Exit;
  result:= jhb;
end;

procedure TParametrizationR2R4.PosFirstAndSecondDerivativesFR2R4(u, v:
  Double;
  var y, yu, yv, yuu, yuv, yvv: Vector4);
  {gives the value of the map at (u,v) and its first and second derivatives}
var
  fd: SecondDeriv;
  i: Integer;
begin
  for i:= 1 to 4 do begin
    fd:= theFs[i].SecondD([u, v, 1, 0, 0, 0]);
    y[i]:= fd.fa0;
    yu[i]:= fd.dfa0;
    yuu[i]:= fd.ddfa0;
    fd:= theFs[i].SecondD([u, v, 0, 1, 0, 0]);
    yv[i]:= fd.dfa0;
    yvv[i]:= fd.ddfa0;
    fd:= theFs[i].SecondD([u, v, 1, 1, 0, 0]);
    yuv[i]:= 0.5 * (fd.ddfa0 - yuu[i] - yvv[i]);
  end;
end;
procedure TParametrizationR2R4.PosFirstDerivativesFR2R4(u, v: Double;
  var y, yu, yv: Vector4);
{gives the value of the map at (u,v) and its first derivatives}
var
  fd: FirstDeriv;
  i: Integer;
begin
  for i:= 1 to 4 do begin
    fd:= theFs[i].FirstD([u, v, 1, 0]);
    y[i]:= fd.fa0;
    yu[i]:= fd.dfa0;
    fd:= theFs[i].FirstD([u, v, 0, 1]);
    yv[i]:= fd.dfa0;
  end;
end;

procedure TParametrizationR2R4.PosFirstSecondThirdDerivativesFR2R4(u, v:
  Double);
{gives the value of the map at (u,v) and its first, second and third derivatives}
var
  x: Vector4;
  i: Integer;
begin
  PosFirstAndSecondDerivativesFR2R4(u, v, x, xu, xv, xuu, xuv, xvv);
  for i := 1 to 4 do begin
     xuuu[i]:= theFPD[i].ff[1,1,1]([u,v]);
     xuuv[i]:= theFPD[i].ff[1,1,2]([u,v]);
     xuvv[i]:= theFPD[i].ff[1,2,2]([u,v]);
     xvvv[i]:= theFPD[i].ff[2,2,2]([u,v]);
  end;
 end;

initialization
  Initialize;

finalization
  Finalize;

end.

(*   CORRECT BUT NOT USED


function TParametrizationR2R4.ComputePQR(x, y: Double; var p, q, r, a, b, c:
  Double): Boolean;
{Stores all basis data of the first and second fundamental forms and gives also
 the numbers 'p,q' such that H = p B+ q C; the number  r= sqrt(x^2+y^2), and the
 numbers 'a,b,c' such that if (v1, v2) are the components of a tangent vector in
 the orthonormal basis that we use, then its components in the coordinate basis are
 c1= a*v1 + b* v2,   c2= c*v2}
var
  pos: Vector4;
  G11, G12, G21, G22, vH, vB, vC: Vector4;
  aux, auxbis: Double;
  upuu, upuv, upvu, upvv, vpuu, vpuv, vpvu, vpvv: Double;
  i: Integer;
begin
  Result:= False;
  PosFirstAndSecondDerivativesFR2R4(x, y, pos, xu, xv, xuu, xuv, xvv);
  guu:= EuclInnerProd(xu, xu);
  guv:= EuclInnerProd(xu, xv);
  gvv:= EuclInnerProd(xv, xv);
  auxbis:= guu * gvv - sqr(guv);
  try
    if auxbis > 1.0E-10 then begin
      aux:= 1 / sqrt(guu);
      a:= aux;
      auxbis:= aux / sqrt(auxbis);
      b:= -guv * auxbis;
      c:= guu * auxbis;
      for i:= 1 to 4 do begin
        t[i]:= xu[i] * aux;
        Jt[i]:= auxbis * (guu * xv[i] - guv * xu[i]);
        G21[i]:= guu * xuv[i] - guv * xuu[i];
      end;
    end else
      Exit;
    G11:= xuu;
    upuu:= EuclInnerProd(xu, xuu);
    upuv:= EuclInnerProd(xu, xuv);
    upvv:= EuclInnerProd(xu, xvv);
    vpuu:= EuclInnerProd(xv, xuu);
    vpuv:= EuclInnerProd(xv, xuv);
    for i:= 1 to 4 do begin
      G12[i]:= guu * xuv[i] - guv * xuu[i] + 2 * upuu * xv[i] - (vpuu + upuv) *
        xu[i];
      G22[i]:= guu * (guu * xvv[i] - guv * xuv[i] + 2 * upuv * xv[i] - (vpuv +
        upvv)
        * xu[i]) - guv * G12[i];
    end;
    upuu:= upuu * aux;
    vpuu:= EuclInnerProd(Jt, G11);
    upuv:= EuclInnerProd(t, G12);
    vpuv:= EuclInnerProd(Jt, G12);
    upvu:= EuclInnerProd(t, G21);
    vpvu:= EuclInnerProd(Jt, G21);
    upvv:= EuclInnerProd(t, G22);
    vpvv:= EuclInnerProd(Jt, G22);
    aux:= aux * auxbis;
    auxbis:= sqr(auxbis);
    for i:= 1 to 4 do begin
      G11[i]:= (G11[i] - upuu * t[i] - vpuu * Jt[i]) / guu;
      G21[i]:= (G21[i] - upvu * t[i] - vpvu * Jt[i]) * aux;
      G12[i]:= (G12[i] - upuv * t[i] - vpuv * Jt[i]) * aux;
      G22[i]:= (G22[i] - upvv * t[i] - vpvv * Jt[i]) * auxbis;
      vH[i]:= (G11[i] + G22[i]) / 2;
      vB[i]:= (G11[i] - G22[i]) / 2;
      vC[i]:= (G12[i] + G21[i]) / 2;
    end;
    hh:= EuclInnerProd(vH, vH);
    nh:= sqrt(hh);
    bb:= EuclInnerProd(vB, vB);
    nb:= sqrt(bb);
    cc:= EuclInnerProd(vC, vC);
    nc:= sqrt(cc);
    bc:= EuclInnerProd(vB, vC);
    hb:= EuclInnerProd(vH, vB);
    hc:= EuclInnerProd(vH, vC);
    aux:= bb * cc - sqr(bc);
    if aux < 1.0E-13 then
      Exit;
    p:= (hb * cc - hc * bc) / aux;
    q:= (hc * bb - hb * bc) / aux;
    r:= hypot(p, q);
    result:= (r > 1 + 1.0e-12);
  except on E: Exception do
      result:= False;
  end;
end;   *)







{   NOT USED


function TParametrizationR2R4.ComputeCuspFromDerivatives: Double;
var
  aux, a, d, p, q, r, alpha, beta, co, se, dif: Double;
  nu, nv, nuu, nuv, nvv, nuuu, nuvv, nuuv, nvvv, up, vp, us, vs, ut, vt:
    Double;
  i: Integer;
  m: matrix44;
begin
  Result:= kTheError;
  aux:= bb * cc - sqr(bc);
  if aux < 1.0E-13 then
    Exit;
  p:= (hb * cc - hc * bc) / aux;
  q:= (hc * bb - hb * bc) / aux;
  alpha:= MyArctan2(-p, q);
  r:= hypot(p, q);
  if r < 1 then
    Exit;
  beta:= arcsin(1 / r);
  dif:= (alpha - beta) / 2;
  beta:= (alpha + beta + pi) / 2;
  alpha:= dif;
  co:= cos(alpha);
  se:= sin(alpha);
  for i:= 1 to 4 do
    m[1, i]:= Jt[i] * co - t[i] * se;
  nu:= EuclInnerProd(m[1], xu);
  nv:= EuclInnerProd(m[1], xv);
  nuu:= EuclInnerProd(m[1], xuu);
  nuv:= EuclInnerProd(m[1], xuv);
  nvv:= EuclInnerProd(m[1], xvv);
  nuuu:= EuclInnerProd(m[1], xuuu);
  nuuv:= EuclInnerProd(m[1], xuuv);
  nuvv:= EuclInnerProd(m[1], xuvv);
  nvvv:= EuclInnerProd(m[1], xvvv);
  up:= -nv;
  vp:= nu;
  us:= -(nuv * up + nvv * vp);
  vs:= nuu * up + nuv * vp;
  ut:= -(nuuv * sqr(up) + 2 * nuvv * up * vp + nvvv * sqr(vp) + nuv * us + nvv
    * vs);
  vt:= nuuu * sqr(up) + 2 * nuuv * up * vp + nuvv * sqr(vp) + nuu * us + nuv *
    vs;
  for i:= 1 to 4 do begin
    m[2, i]:= xu[i] * up + xv[i] * vp;
    m[3, i]:= xuu[i] * sqr(up) + 2 * xuv[i] * up * vp + xvv[i] * sqr(vp) +
      xu[i]
      * us + xv[i] * vs;
    m[4, i]:= xu[i] * ut + xv[i] * vt + 3 * (xuu[i] * up * us + xvv[i] * vp *
      vs
      + xuv[i] * (up * vs + vp * us));
    m[4, i]:= m[4, i] + xuuu[i] * up * sqr(up) + xvvv[i] * vp * sqr(vp) + 3 *
      (xuuv[i] * sqr(up) * vp + xuvv[i] * sqr(vp) * up);
  end;
  Determinant(d, m);
  aux:= (sqr(m[2, 1]) + sqr(m[2, 2]) + sqr(m[2, 3]) + sqr(m[2, 4])) * (sqr(m[3,
    1]) + sqr(m[3, 2]) + sqr(m[3, 3]) + sqr(m[3, 4]));
  aux:= aux - sqr(m[2, 1] * m[3, 1] + m[2, 2] * m[3, 2] + m[2, 3] * m[3, 3] +
    m[2, 4] * m[3, 4]);
  if aux < 1.0E-13 then
    Exit;
  d:= d / aux;

  alpha:= beta;
  co:= cos(alpha);
  se:= sin(alpha);
  for i:= 1 to 4 do
    m[1, i]:= Jt[i] * co - t[i] * se;
  nu:= EuclInnerProd(m[1], xu);
  nv:= EuclInnerProd(m[1], xv);
  nuu:= EuclInnerProd(m[1], xuu);
  nuv:= EuclInnerProd(m[1], xuv);
  nvv:= EuclInnerProd(m[1], xvv);
  nuuu:= EuclInnerProd(m[1], xuuu);
  nuuv:= EuclInnerProd(m[1], xuuv);
  nuvv:= EuclInnerProd(m[1], xuvv);
  nvvv:= EuclInnerProd(m[1], xvvv);
  up:= -nv;
  vp:= nu;
  us:= -(nuv * up + nvv * vp);
  vs:= nuu * up + nuv * vp;
  ut:= -(nuuv * sqr(up) + 2 * nuvv * up * vp + nvvv * sqr(vp) + nuv * us + nvv
    * vs);
  vt:= nuuu * sqr(up) + 2 * nuuv * up * vp + nuvv * sqr(vp) + nuu * us + nuv *
    vs;
  for i:= 1 to 4 do begin
    m[2, i]:= xu[i] * up + xv[i] * vp;
    m[3, i]:= xuu[i] * sqr(up) + 2 * xuv[i] * up * vp + xvv[i] * sqr(vp) +
      xu[i]
      * us + xv[i] * vs;
    m[4, i]:= xu[i] * ut + xv[i] * vt + 3 * (xuu[i] * up * us + xvv[i] * vp *
      vs
      + xuv[i] * (up * vs + vp * us));
    m[4, i]:= m[4, i] + xuuu[i] * up * sqr(up) + xvvv[i] * vp * sqr(vp) + 3 *
      (xuuv[i] * sqr(up) * vp + xuvv[i] * sqr(vp) * up);
  end;
  Determinant(a, m);
  aux:= (sqr(m[2, 1]) + sqr(m[2, 2]) + sqr(m[2, 3]) + sqr(m[2, 4])) * (sqr(m[3,
    1]) + sqr(m[3, 2]) + sqr(m[3, 3]) + sqr(m[3, 4]));
  aux:= aux - sqr(m[2, 1] * m[3, 1] + m[2, 2] * m[3, 2] + m[2, 3] * m[3, 3] +
    m[2, 4] * m[3, 4]);
  if aux < 1.0E-13 then
    Exit;
  result:= d * a / aux;
end;}

{ NOT USED

function TParametrizationR2R4.ComputeCuspFunct(const p: Vector2): Double;
begin
  Result:= kTheError;
  if not ComputeHBCandProducts(p) then
    Exit;
  Result:= ComputeCuspFromDerivatives;
end;}



(*   NOT USED

function TParametrizationR2R4.ComputeInvariantsAndCusp(const par: vector2; var
  halfMajorAxis, halfMinorAxis, longH, depth, alpha, KGAuss, discrNormDev, discrLatDev: Double): Boolean;
{Nos da, en el punto de la superficie x(p) el valor de los ejes de la elipse de curvatura,}
{la longitud 'longH' de la curvatura media, y en 'depth', cunto est el  0 del tangente}
{dentro de la  elipse de curvatura -valores positivos de depth- o fuera -valores negativos-.}
var
  phi_M, a, b, g, aux, si, co: Double;
  m: matrix44;
  axis: Vector4;
  i: Integer;
begin
  Result:= False;
  try
    if not ComputeHBCandProducts(par) then
      Exit;
    discrNormDev:= DiscriminantExtremalNormDevDirs(par);
    discrLatDev:= DiscriminantExtremalLatDevDirs(par);
    KGAuss:= hh - bb - cc;
    longH:= sqrt(hh);
    aux:= bb * cc - sqr(bc);
    if abs(aux) < 1.0E-12 then
      depth:= -11
    else begin
      a:= (hb * cc - hc * bc);
      b:= (hc * bb - hb * bc);
      depth:= 1 - hypot(a, b) / aux;
    end;
    aux:= hypot(bb - cc, 2 * bc);
    g:= bb + cc;
    halfMajorAxis:= sqrt(abs(g + aux) / 2);
    halfMinorAxis:= sqrt(abs(g - aux) / 2);

    if (aux < 1.0E-8) then
      alpha:= kTheError
    else begin
      phi_M:= MyArctan2(2 * bc, bb - cc) / 2;
      sincos(phi_M, si, co);
      aux:= 0;
      for i:= 1 to 4 do begin
        axis[i]:= vB[i] * co + vC[i] * si;
        aux:= aux + sqr(axis[i]);
      end;
      m[1]:= axis;
      m[2]:= t;
      m[3]:= Jt;
      m[4]:= vH;
      Determinant(alpha, m);
      aux:= sqrt(aux) * nh;
      if aux > 1.0e-12 then
        alpha:= abs(alpha / aux)
      else
        alpha:= kTheError;
    end;
    Result:= True;
  except on E: Exception do
      result:= False;
  end;
end;   *)




